"""
backend.py
============

Contains implementations for different backends where data may be registered

"""
import os
import re
import sys
import logging
from typing import List
import json

import django
from django.db import transaction
from django.db.models import Count, Q
from django.contrib.gis.geos import GEOSGeometry
from osgeo import gdal

from .exceptions import RegistrationError
from .context import Context
from .json import read_json
from .source import Source, LocalSource, S3Source, SwiftSource
from .utils import import_by_path


logger = logging.getLogger(__name__)


class RegistrationResult:
    pass


class Backend:
    def register(self, item: Context) -> RegistrationResult:
        raise NotImplementedError

    def deregister(self, item: Context) -> List[str]:
        return []


class EOxServerBackend(Backend):
    """EOxServer backend allows registration to be performed on a running EOxServer instance

    Args:
        instance_base_path (str): base path of the instance
        instance_name (str): name of django instance
        mapping (dict): mapping dictionary of values between backend and 
        simplify_footprint_tolerance (int, optional): [description]. Defaults to None.
    """
    def __init__(self, instance_base_path: str, instance_name: str,
                 mapping: dict, simplify_footprint_tolerance: int=None):
        self.mapping = mapping
        self.simplify_footprint_tolerance = simplify_footprint_tolerance
        path = os.path.join(instance_base_path, instance_name)
        if path not in sys.path:
            sys.path.append(path)

        os.environ.setdefault(
            "DJANGO_SETTINGS_MODULE",
            f"{instance_name}.settings"
        )
        django.setup()

    def exists(self, source: Source, item: Context):
        """Checks whether the item exists in the given source

        Args:
            source (Source): source of the data
            item (Context): item to be checked

        Returns:
            (bool): true if exists, false otherwise
        """
        from eoxserver.resources.coverages import models
        return models.Product.objects.filter(
            identifier=item.identifier
        ).exists()

    def _get_storage_from_source(self, source: Source, path: str) -> list:
        from eoxserver.backends import models as backends

        created_storage_auth = False
        created_storage = False
        storage_name = None
        if isinstance(source, LocalSource):
            storage, created_storage = backends.Storage.get_or_create(
                name=source.name,
                url=source.root_directory,
                storage_type='local',
            )
            storage_name = storage.name

        elif isinstance(source, S3Source):
            params = json.dumps({
                'ACCESS_KEY_ID': source.access_key_id,
                'SECRET_ACCESS_KEY': source.secret_access_key,
                'AWS_REGION': source.region_name
            })

            endpoint_url = source.endpoint_url
            if endpoint_url.startswith('https://'):
                endpoint_url = endpoint_url[len('https://'):]
            elif endpoint_url.startswith('http://'):
                endpoint_url = endpoint_url[len('http://'):]

            bucket, _ = source.get_container_and_path(path)

            storage_auth, created_storage_auth = backends.StorageAuth.objects.get_or_create(
                name=source.name if source.bucket_name else f'{source.name}-{bucket}',
                url=endpoint_url,
                storage_auth_type='S3',
            )

            storage_auth.auth_parameters = params
            storage_auth.save()

            storage, created_storage = backends.Storage.objects.get_or_create(
                name=source.name if source.bucket_name else f'{source.name}-{bucket}',
                url=bucket,
                storage_type='S3',
                storage_auth=storage_auth,
            )
            storage_name = storage.name

        elif isinstance(source, SwiftSource):
            params = json.dumps({
                'auth-version': str(source.auth_version),
                'identity-api-version': str(source.auth_version),
                'username': source.username,
                'password': source.password,
                'tenant-name': source.tenant_name,
                'tenant-id': source.tenant_id,
                'region-name': source.region_name,
            })

            storage_auth, created_storage_auth = backends.StorageAuth.objects.get_or_create(
                name=source.auth_url,
                url=source.auth_url_short or source.auth_url,
                storage_auth_type='keystone',
                auth_parameters=params,
            )

            container, _ = source.get_container_and_path(path)

            storage, created_storage = backends.Storage.objects.get_or_create(
                name=source.name if source.container else f'{source.name}{container}',
                url=container,
                storage_type='swift',
                storage_auth=storage_auth,
            )
            storage_name = storage.name

        if created_storage_auth:
            logger.info(f'Created storage auth for {source.name}')
        if created_storage:
            logger.info(f'Created storage for {source.name}')

        return [storage_name] if storage_name else []

    def _register_with_registrator(self, source: Source, item: Context, replace: bool, storage, mapping) -> RegistrationResult:
        from eoxserver.resources.coverages import models
        from eoxserver.resources.coverages.registration.product import ProductRegistrator
        from eoxserver.resources.coverages.registration.browse import BrowseRegistrator
        from eoxserver.resources.coverages.registration.mask import MaskRegistrator
        from eoxserver.resources.coverages.registration.registrators.gdal import GDALRegistrator

        if not mapping:
            raise RegistrationError(f'Could not get mapping for {item.product_type} {item.product_level}')

        try:
            _, metadata_file = source.get_container_and_path(item.metadata_files[0])
            metadata_locations = [storage + [metadata_file]]
        except IndexError:
            metadata_locations = []

        try:
            models.ProductType.objects.get(name=mapping['product_type_name'])
        except models.ProductType.DoesNotExist:
            pass

        if 'footprint' in item.metadata:
            footprint = GEOSGeometry(item.metadata.pop('footprint'))
        else:
            footprint = None

        product, _ = ProductRegistrator().register(
            metadata_locations=metadata_locations,
            type_name=mapping['product_type_name'],
            replace=replace,
            extended_metadata=True,
            mask_locations=None,
            package_path=None,
            simplify_footprint_tolerance=self.simplify_footprint_tolerance,
            overrides=dict(
                identifier=item.identifier,
                footprint=footprint,
                **item.metadata
            ),
        )
        if not product.footprint or product.footprint.empty:
            raise RegistrationError("No footprint was extracted. full product: %s" % product)

        # register coverages and link them to the product
        for raster_identifier, coverage_type_name in mapping.get('coverages', {}).items():
            raster_items = item.raster_files.get(raster_identifier)
            raster_items = [
                storage + [source.get_container_and_path(raster_item)[1]]
                for raster_item in (raster_items if isinstance(raster_items, list) else [raster_items])
            ]
            if isinstance(coverage_type_name, list):
                # determine which one to use based on number of bands
                _, vsi_path = source.get_vsi_env_and_path(item.raster_files.get(raster_identifier)[0])
                ds = gdal.Open(vsi_path)
                logger.debug("Raster has number of bands: '%s'" % ds.RasterCount)
                # find covtype that has correct number of bands
                try:
                    coverage_model = models.CoverageType.objects \
                        .filter(name__in=coverage_type_name) \
                        .annotate(field_count=Count('field_types')) \
                        .get(field_count=ds.RasterCount)
                except models.CoverageType.DoesNotExist:
                    raise RegistrationError(f'Coverage Band Count is different than in any of configured CoverageTypes')
                coverage_type_name = coverage_model.name
            logger.info(f"Registering coverage{'s' if len(raster_items) > 1 else ''} {raster_items} as {coverage_type_name}")
            report = GDALRegistrator().register(
                data_locations=raster_items,
                metadata_locations=metadata_locations,
                coverage_type_name=coverage_type_name,
                overrides={
                    "identifier": f'{product.identifier}__{raster_identifier}__coverage',
                    "footprint": None,
                },
                replace=replace,
            )
            logger.debug("Adding coverage to product")
            models.product_add_coverage(product, report.coverage)

        # register browses
        for raster_identifier, browse_type_name in (mapping.get('browses') or {}).items():
            raster_item = item.raster_files.get(raster_identifier)

            _, raster_item = source.get_container_and_path(raster_item)
            logger.info(f"Adding browse {browse_type_name or 'default'} {raster_item} to product")

            BrowseRegistrator().register(
                product.identifier,
                storage + [raster_item],
                browse_type_name,
            )

        # register masks
        for mask_identifier, mask_type_name in (mapping.get('masks') or {}).items():
            mask_item = item.mask_files.get(mask_identifier)
            if mask_item:
                _, mask_path = source.get_container_and_path()
                logger.info(f"Adding mask (file) {mask_type_name} to product")
                MaskRegistrator().register(
                    product.identifier,
                    storage + [mask_path],
                    mask_type_name,
                )

            mask_item = item.masks.get(mask_identifier)
            if mask_item:
                logger.info(f"Adding mask (geometry) {mask_type_name} to product")
                models.Mask.objects.create(
                    product=product,
                    mask_type=models.MaskType.objects.get(
                        product_type=product.product_type,
                        name=mask_type_name,
                    ),
                    geometry=mask_item,
                )

        return product

    def _register_with_stac(self, source: Source, item: Context, replace: bool, storage, mapping: dict) -> RegistrationResult:
        from eoxserver.backends import models as backends
        from eoxserver.resources.coverages.registration.stac import (
            register_stac_product, create_product_type_from_stac_item
        )

        stac_item = read_json(source, item.path)

        if not mapping:
            logger.info(
                f'No mapping available, creating new Product Type named {item.product_type}'
            )
            product_type, _ = create_product_type_from_stac_item(
                stac_item,
                product_type_name=item.product_type,
                ignore_existing=True,
                coverage_mapping=None
            )
            logger.info(f'Created Product Type {product_type.name}')

        elif not item.product_type:
            logger.info(
                'No mapping available, creating new Product Type with generated name'
            )
            product_type, _ = create_product_type_from_stac_item(
                stac_item,
                product_type_name=mapping.get('product_type_name'),
                ignore_existing=True,
                coverage_mapping=mapping.get('coverages', {})
            )
            logger.info(f'Created Product Type {product_type.name}')

        else:
            try:
                product_type = mapping['product_type_name']
                logger.info(f'Using Product Type {product_type}')
            except KeyError:
                raise RegistrationError('Mapping for product_type_name not set')

        # resolve storage object
        if storage:
            storage = backends.Storage.objects.get(name=storage[0])

        _, path = source.get_container_and_path(item.path)
        product, _ = register_stac_product(
            path,
            stac_item,
            product_type,
            storage,
            replace=replace,
            mapping=mapping.get('coverages', {})
        )
        return product

    @transaction.atomic
    def register(self, source: Source, item: Context, replace: bool) -> RegistrationResult:
        """Registers the item to the endpoint

        Args:
            source (Source): source of the data
            item (Context): item to be registered
            replace (bool): replace existing or not

        """
        # ugly, ugly hack
        from eoxserver.resources.coverages import models

        storage = self._get_storage_from_source(source, item.path)

        logger.info(
            f'Registering into EOxServer for type "{item.product_type}" '
            f'and level "{item.product_level}".'
        )

        product_type = item.product_type or ''

        # get the mapping for this particular item
        mapping = {}
        try:
            type_mapping = self.mapping[product_type]
            logger.info(f'Using Product Type mapping {product_type}')
            product_level = item.product_level or ''
            if product_level in type_mapping:
                mapping = type_mapping[product_level]
                logger.info(f'Using Product Level mapping {product_level}')
            elif None in type_mapping:
                mapping = type_mapping[None]
                logger.info('Using default level mapping')
        except KeyError:
            pass

        logger.info(f'Final mapping: {mapping}')

        # branch here for STAC Item registration, as a product type can be
        # dynamically generated
        if item.scheme == 'stac-item':
            product = self._register_with_stac(source, item, replace, storage, mapping)
        else:
            product = self._register_with_registrator(source, item, replace, storage, mapping)

        # insert the product in the to be associated collections
        for collection_id in mapping.get('collections', []):
            collection = models.Collection.objects.get(identifier=collection_id)
            models.collection_insert_eo_object(collection, product)

    @transaction.atomic
    def deregister(self, item: Context):
        """Attempts to deregister item

        Args:
            item (Context): item to be deregistered
        """
        # ugly, ugly hack
        from eoxserver.resources.coverages import models
        try:
            logger.info(f"Deregistering product '{item.identifier}'")
            product = models.Product.objects.get(identifier=item.identifier)
            grids = list(models.Grid.objects.filter(
                coverage__parent_product=product
            ))
            product.delete()

            # clean up grids
            for grid in grids:
                grid_used = models.EOObject.objects.filter(
                    Q(coverage__grid=grid) | Q(mosaic__grid=grid),
                ).exists()
                # clean up grid as well, if it is not referenced
                # anymore but saving named (user defined) grids
                if grid and not grid.name and not grid_used:
                    grid.delete()
        except models.Product.DoesNotExist:
            logger.info(
                f"No product with identifier '{item.identifier}' found"
            )
            # no product found with that id
            # return empty list
            return []

        logger.info(
            f"Deregistered product with identifier '{item.identifier}'"
        )
        # return the deleted identifier
        return [item.identifier]


BACKENDS = {
    'eoxserver': EOxServerBackend,
}


def get_backends(config: dict, context: Context) -> List[Backend]:
    """Gets the backends for registering the data into

    Args:
        config (dict): registrar configuration
        context (Context): context to register

    Raises:
        RegistrationError: raised if no suitable backend is found for the context

    Returns:
        List[Backend]: list of backends where registration will be carried out
    """
    backends = []
    for cfg_backend in config['backends']:
        schemes = cfg_backend.get('schemes')
        filter_ = cfg_backend.get('filter')
        # filter by scheme/path
        if context.scheme is not None:
            if schemes and context.scheme not in schemes:
                continue

        if context.path is not None:
            if filter_ and not re.match(filter_, context.path):
                continue

        # construct backend
        if 'type' in cfg_backend:
            backend_cls = BACKENDS[cfg_backend['type']]
        else:
            backend_cls = import_by_path(cfg_backend['path'])

        backends.append(
            backend_cls(
                *cfg_backend.get('args', []),
                **cfg_backend.get('kwargs', {}),
            )
        )

    if not backends:
        raise RegistrationError(
            f'Could not find a suitable backend for the path {context.path}'
        )

    return backends
