"""
scheme.py
================

Contains classes for different registration schemes
"""
import re
from os.path import join, normpath, dirname
import logging
from typing import List, Tuple
from datetime import datetime

from .xml import read_xml, parse_metadata_schema, Parameter
from .json import read_json
from .context import Context
from .source import Source
from .exceptions import RegistrationError
from .utils import isoformat, import_by_path, extract_footprint, extract_metadata

logger = logging.getLogger(__name__)


class RegistrationScheme:
    def get_context(self, source: Source, path: str) -> List[Context]:
        raise NotImplementedError


def parse_datetime(value):
    return value


def pairwise(iterable):
    "s -> (s0,s1), (s2,s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)


def parse_footprint(value):
    coord_list = ','.join(
        f'{x} {y}'
        for y, x in pairwise(value.split())
    )
    return f'POLYGON(({coord_list}))'


class Sentinel2RegistrationScheme(RegistrationScheme):
    name = 'sentinel-2'

    MTD_SCHEMAS = {
        'MSIL1C': {
            'begin_time': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/PRODUCT_START_TIME/text()', False, parse_datetime),
            'end_time': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/PRODUCT_STOP_TIME/text()', False, parse_datetime),
            'identifier': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/PRODUCT_URI/text()'),
            'footprint': Parameter('/n1:Level-1C_User_Product/n1:Geometric_Info/Product_Footprint/Product_Footprint/Global_Footprint/EXT_POS_LIST/text()', False, parse_footprint),
            'level': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/PROCESSING_LEVEL/text()'),
            'type': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/PRODUCT_TYPE/text()'),
            'generation_time': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/GENERATION_TIME/text()', False, parse_datetime),
            'cloud_cover': Parameter('/n1:Level-1C_User_Product/n1:Quality_Indicators_Info/Cloud_Coverage_Assessment/text()'),
            'image_file_paths': Parameter('/n1:Level-1C_User_Product/n1:General_Info/Product_Info/Product_Organisation/Granule_List/Granule/IMAGE_FILE/text()', True),
        },
        'MSIL2A': {
            'begin_time': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/PRODUCT_START_TIME/text()', False, parse_datetime),
            'end_time': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/PRODUCT_STOP_TIME/text()', False, parse_datetime),
            'identifier': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/PRODUCT_URI/text()'),
            'footprint': Parameter('/n1:Level-2A_User_Product/n1:Geometric_Info/Product_Footprint/Product_Footprint/Global_Footprint/EXT_POS_LIST/text()', False, parse_footprint),
            'level': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/PROCESSING_LEVEL/text()'),
            'type': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/PRODUCT_TYPE/text()'),
            'generation_time': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/GENERATION_TIME/text()', False, parse_datetime),
            'cloud_cover': Parameter('/n1:Level-2A_User_Product/n1:Quality_Indicators_Info/Cloud_Coverage_Assessment/text()'),
            'image_file_paths': Parameter('/n1:Level-2A_User_Product/n1:General_Info/Product_Info/Product_Organisation/Granule_List/Granule/IMAGE_FILE/text()', True),
        },
    }

    MTD_TL_SCHEMAS = {
        'MSIL1C': {
            'mask_file_paths': Parameter('/n1:Level-2A_Tile_ID/n1:Quality_Indicators_Info/Pixel_Level_QI/MASK_FILENAME/text()', True),
        },
        'MSIL2A': {
            'mask_file_paths': Parameter('/n1:Level-2A_Tile_ID/n1:Quality_Indicators_Info/Pixel_Level_QI/MASK_FILENAME/text()', True),
        },
    }

    MTD_NAMESPACES = {
        'MSIL1C': {
            'n1': "https://psd-14.sentinel2.eo.esa.int/PSD/User_Product_Level-1C.xsd"
        },
        'MSIL2A': {
            'n1': "https://psd-14.sentinel2.eo.esa.int/PSD/User_Product_Level-2A.xsd"
        },
    }

    MTD_TL_NAMESPACES = {
        'MSIL1C': {
            'n1': 'https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-1C_Tile_Metadata.xsd'
        },
        'MSIL2A': {
            'n1': 'https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-2A_Tile_Metadata.xsd'
        },
    }

    def get_context(self, source: Source, path: str) -> List[Context]:
        try:
            metadata_file = join(path, 'MTD_MSIL2A.xml')
            tree = read_xml(source, metadata_file)
            mtd_schema = self.MTD_SCHEMAS['MSIL2A']
            mtd_namespaces = self.MTD_NAMESPACES['MSIL2A']
            mtd_tl_schema = self.MTD_TL_SCHEMAS['MSIL2A']
            mtd_tl_namespaces = self.MTD_TL_NAMESPACES['MSIL2A']
            band_re = re.compile(r'.*([A-Z0-9]{3}_[0-9]{2}m)$')

        except Exception:
            metadata_file = join(path, 'MTD_MSIL1C.xml')
            tree = read_xml(source, metadata_file)
            mtd_schema = self.MTD_SCHEMAS['MSIL1C']
            mtd_namespaces = self.MTD_NAMESPACES['MSIL1C']
            mtd_tl_schema = self.MTD_TL_SCHEMAS['MSIL1C']
            mtd_tl_namespaces = self.MTD_TL_NAMESPACES['MSIL1C']
            band_re = re.compile(r'.*([A-Z0-9]{3})$')

        # get product metadata
        metadata = parse_metadata_schema(tree, mtd_schema, mtd_namespaces)
        raster_files = {
            band_re.match(image_file_path).groups()[0]: f'{join(path, image_file_path)}.jp2'
            for image_file_path in metadata['image_file_paths']
        }

        # get granule metadata
        mtd_files = source.list_files(join(path, 'GRANULE'), '*/MTD_TL.xml')
        logger.info(f'{mtd_files}')
        tl_tree = read_xml(source, mtd_files[0])
        tile_metadata = parse_metadata_schema(tl_tree, mtd_tl_schema, mtd_tl_namespaces)

        mask_type_re = re.compile(r'.*/MSK_([A-Z]*)_([A-Z0-9]{3}).[a-z0-9]+$')
        mask_files = {
            mask_type_re.match(mask_file_path).groups()[0]: join(path, mask_file_path)
            for mask_file_path in tile_metadata['mask_file_paths']
            if mask_type_re.match(mask_file_path) is not None
        }

        return [Context(
            identifier=metadata['identifier'],
            path=path,
            scheme=self.name,
            product_type=metadata['type'],
            product_level=metadata['level'],
            raster_files=raster_files,
            mask_files=mask_files,
            metadata_files=[metadata_file],
            metadata={
                'bclass registrar.scheme.SingleFileRegistrationScheme(product_type: str, level: str, date_format: str, id_template: Optional[str] = None, date_regex: Optional[str] = None, footprint: Optional[str] = Nonegin_time': metadata['begin_time'],
                'end_time': metadata['end_time'],
                'generation_time': metadata['generation_time'],
                'cloud_cover': metadata['cloud_cover'],
                'footprint': metadata['footprint'],
            }
        )]


def parse_ring(string):
    raw_coords = string.split()
    return [(lon, lat) for lat, lon in pairwise(raw_coords)]


def parse_polygons_gsc(elem):
    def serialize_coord_list(coords):
        return ','.join(
            f'{x} {y}' for x, y in coords
        )

    interior = serialize_coord_list(
        parse_ring(
            elem.xpath(
                "gml:exterior/gml:LinearRing/gml:posList", namespaces=elem.nsmap
            )[0].text.strip()
        )
    )

    exteriors = [
        f'''({
            serialize_coord_list(
                parse_ring(poslist_elem.text.strip())
            )
        })'''
        for poslist_elem in elem.xpath(
            "gml:interior/gml:LinearRing/gml:posList", namespaces=elem.nsmap
        )
    ]

    return f"POLYGON(({interior}){',' if exteriors else ''}{','.join(exteriors)})"


class GSCRegistrationScheme(RegistrationScheme):
    """Used to register GSC data

    Args:
        level_re (str, optional): regular expression of the level. Defaults to r'.*(Level_[0-9]+)$'.
        product_type_mapping(str, optional): xml extractor for product type. Defaults to '//gml:using/eop:EarthObservationEquipment/eop:platform/eop:Platform/eop:shortName/text()'.
    """
    name = 'gsc'

    GSC_SCHEMA = {
        'identifier': Parameter('//gml:metaDataProperty/gsc:EarthObservationMetaData/eop:identifier/text()'),
        'level': Parameter('//gml:metaDataProperty/gsc:EarthObservationMetaData/eop:parentIdentifier/text()'),
        'mask': Parameter('//gsc:opt_metadata/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:vendorSpecific/eop:SpecificInformation[eop:localAttribute/text() = "CF_POLY"]/eop:localValue/text()', True),
        'footprint': Parameter(
            '//gml:target/eop:Footprint/gml:multiExtentOf/gml:MultiSurface/gml:surfaceMembers/gml:Polygon',
            False, parse_polygons_gsc
        ),
    }

    def __init__(self, level_re: str=r'.*(Level_[0-9]+)$', product_type_mapping: str='//gml:using/eop:EarthObservationEquipment/eop:platform/eop:Platform/eop:shortName/text()'):
        self.level_re = level_re
        self.GSC_SCHEMA['type'] = Parameter(product_type_mapping)

    def get_context(self, source: Source, path: str) -> List[Context]:
        gsc_filenames = source.list_files(path, ['GSC*.xml', 'GSC*.XML'])
        metadata_file = gsc_filenames[0]

        tree = read_xml(source, metadata_file)
        metadata = parse_metadata_schema(
            tree, self.GSC_SCHEMA, tree.getroot().nsmap
        )

        tiff_files = {
            metadata['type']: source.list_files(path, ['*.tif', '*.TIF'])
        }

        match = re.match(self.level_re, metadata['level'])
        if match:
            level = match.groups()[0]
        else:
            level = None

        return [Context(
            identifier=metadata['identifier'],
            path=path,
            scheme=self.name,
            product_type=metadata['type'],
            product_level=level,
            raster_files=tiff_files,
            masks={
                'validity': metadata['mask'][0] if metadata['mask'] else None
            },
            metadata_files=[metadata_file],
            metadata={
                'footprint': metadata['footprint'],
            }
        )]


class STACItemScheme(RegistrationScheme):
    """Used to register stac items

    Args:
        product_type_property (str, optional): [description]. Defaults to None.
        product_level_property (str, optional): [description]. Defaults to None.
        product_type (str, optional): [description]. Defaults to None.
        level (str, optional): [description]. Defaults to ''.
        product_type_from_band_assets (bool, optional): [description]. Defaults to False.
    """
    name = 'stac-item'

    def __init__(self,
                 product_type_property: str=None, product_level_property: str=None,
                 product_type: str=None, level: str='',
                 product_type_from_band_assets=False):
        self.product_type_property = product_type_property
        self.product_level_property = product_level_property
        self.product_type_from_band_assets = product_type_from_band_assets
        self.product_type = product_type
        self.level = level

    def get_context(self, source: Source, path: str) -> List[Context]:
        # read json from url or in path
        stac_item = source.list_files(path)

        # get items
        item = read_json(source, stac_item[0])

        # convert items to contexts
        return [
            self._item_to_context(stac_item[0], item)
        ]

    def _item_to_context(self, path, item):
        # treat everything with 'eo:bands' as a valid raster item
        band_assets = {
            asset_name: asset
            for asset_name, asset in item['assets'].items()
            if 'eo:bands' in asset
        }
        raster_files = {
            asset_name: asset['href']
            for asset_name, asset in band_assets.items()
        }

        properties = item.get('properties', {})

        # determine product type:
        if self.product_type:
            # use a fixed product type
            product_type = self.product_type
        if self.product_type_property:
            # use either a specific property
            product_type = properties.get(self.product_type_property)
        elif self.product_type_from_band_assets:
            # get the product type from the band assets names
            product_type = '_'.join(sorted(band_assets.keys()))
        else:
            # use either a specific
            product_type = properties.get('constellation') or properties.get('platform') or item.get('collection') or ''

        if self.product_level_property:
            level = properties.get(self.product_level_property)
        else:
            level = self.level

        return Context(
            identifier=item['id'],
            path=path,
            scheme=self.name,
            product_type=product_type,
            product_level=level,
            metadata=self._transform_metadata(item),
            raster_files=raster_files,
            metadata_files=[path],
            masks=[],
            mask_files=[],
        )

    def _transform_metadata(self, item):
        # TODO: implement
        return {}


class SingleFileRegistrationScheme(RegistrationScheme):
    """Used for registration of single files

    Args:
        product_type (str): type of product
        level (str): level of the product
        date_format (str): format of the date in the file name. Used to extract the date
        id_template (str, optional): Template for the coverage and product id. Defaults to None.
        date_regex (str, optional): regular expression for finding the date in the file name. Defaults to None.
        footprint (str, optional): WKT formatted footprint. Defaults to None.
        id_values (str, optional): regular expression with named groups containing values for the id_template. Defaults to None.
    """
    name = "single-file"

    def __init__(self, product_type: str, level: str,
                 date_format: str, id_template: str=None, date_regex: str=None, footprint: str=None, id_values: str=None):

        self.product_type = product_type
        self.level = level
        self.date_format = date_format
        self.id_template = id_template
        self.id_values = id_values
        self.date_regex = date_regex
        self.footprint = footprint

    def get_context(self, source: Source, path: str) -> List[Context]:

        begin_time, end_time = self._get_datetimes(path)
        identifier = self._get_identifier(path, begin_time, end_time)
        footprint = self._get_footprint(source, path)
        extracted_metadata = self._get_file_metadata(source, path)

        metadata={
            "begin_time": isoformat(begin_time),
            "end_time": isoformat(end_time),
            "footprint": footprint,
        }

        try:
            metadata["cloud_cover"] = extracted_metadata['eo-cloud_cover']
        except KeyError:
            pass

        return [
            Context(
                identifier=identifier,
                path=path,
                scheme=self.name,
                product_type=self.product_type,
                product_level=self.level,
                raster_files={self.name: path},
                mask_files={},
                metadata_files=[],
                metadata=metadata
            )
        ]

    def _get_identifier(self, path:str, begin_time: datetime, end_time: datetime) -> str:
        begin_time=begin_time.strftime("%Y%m%d_%H%M%S")
        end_time=end_time.strftime("%Y%m%d_%H%M%S")
        if not self.id_template:
            return f"{self.product_type}_{begin_time}_{end_time}"

        help_dict  =  {
            'product_type': self.product_type,
            'end_time': end_time
        }

        reg = re.compile(self.id_values)
        id_dict = reg.match(path).groupdict()
        id_dict.update(help_dict)

        return self.id_template.format(**id_dict)



    def _get_datetimes(self, path: str) -> Tuple[datetime, datetime]:
        dt = path.split("/")[-1].split(".")[0]
        dt = re.search(self.date_regex, dt).group()

        begin = datetime.strptime(dt, self.date_format)
        end = begin.replace(hour=23, minute=59, second=59)

        return begin, end

    def _get_footprint(self, source, path):
        if self.footprint:
            return self.footprint
        
        env, vsipath = source.get_vsi_env_and_path(path)
        return extract_footprint(env, vsipath)

    def _get_file_metadata(self, source, path):
        env, vsipath = source.get_vsi_env_and_path(path)
        return extract_metadata(env, vsipath)

REGISTRATION_SCHEMES = {
    'gsc': GSCRegistrationScheme,
    'sentinel-2': Sentinel2RegistrationScheme,
    'stac-item': STACItemScheme,
    'single-file': SingleFileRegistrationScheme,
}


def get_scheme(config: dict, path: str) -> RegistrationScheme:
    """Retrieves scheme based on configuration and path

    Args:
        config (dict): registrar configuration
        path (str): path to register

    Raises:
        RegistrationError: Raised when no suitable scheme is found

    Returns:
        RegistrationScheme: one of REGISTRATION_SCHEMES
    """

    cfg_schemes = config['schemes']

    for cfg_scheme in cfg_schemes:
        if cfg_scheme.get('filter'):
            if re.match(cfg_scheme['filter'], path):
                break
        else:
            break
    else:
        # no source found
        raise RegistrationError(
            f'Could not find a suitable scheme for the path {path}'
        )

    if 'type' in cfg_scheme:
        scheme_cls = REGISTRATION_SCHEMES[cfg_scheme['type']]
    else:
        scheme_cls = import_by_path(cfg_scheme['path'])

    return scheme_cls(
        *cfg_scheme.get('args', []),
        **cfg_scheme.get('kwargs', {}),
    )
