from os.path import join, dirname
import logging.config

import click
import yaml
import jsonschema

from .daemon import run_daemon
from .config import load_config
from .harvester import main
from .utils import get_params

def setup_logging(debug=False):
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'brief': {
                'format': '%(levelname)s %(name)s: %(message)s'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'DEBUG' if debug else 'INFO',
                'formatter': 'brief',
            }
        },
        'root': {
            'handlers': ['console'],
            'level': 'DEBUG' if debug else 'INFO',
        }
    })


def validate_config(config):
    with open(join(dirname(__file__), 'config-schema.yaml')) as f:
        schema = yaml.load(f)

    jsonschema.validate(config, schema)


@click.group()
def cli():
    pass


@cli.command(help='Run the harvester daemon, attaching to a Redis queue')
@click.option('--config-file', type=click.File('r'))
@click.option('--validate/--no-validate', default=False)
@click.option('--host', type=str)
@click.option('--port', type=int)
@click.option('--listen-queue', type=str)
@click.option('--debug/--no-debug', default=False)
def daemon(config_file=None, validate=False, host=None, port=None, listen_queue=None, debug=False):
    setup_logging(debug)
    config = load_config(config_file)
    if validate:
        validate_config(config)
    run_daemon(config, host, port, listen_queue)


@cli.command(help='Run a single, one-off harvest')
@click.argument('harvester_name', type=str)
@click.option('--config-file', type=click.File('r'))
@click.option('--validate/--no-validate', default=False)
@click.option('--debug/--no-debug', default=False)
@click.option( '--param', '-p', multiple=True)
def harvest(harvester_name, config_file: str=None, validate: bool=False, debug: bool=False, param: tuple=()):
    setup_logging(debug)
    config = load_config(config_file)
    if validate:
        validate_config(config)

    kwargs = get_params(param)
    kwargs['name'] = harvester_name
    main(config, harvester_name, **kwargs)

if __name__ == '__main__':
    cli()
