from datetime import datetime
import logging
from typing import Type, List

from pygeofilter.parsers.cql_json import parse as json_parse
from pygeofilter.backends.native.evaluate import NativeEvaluator

from ..query import Query

logger = logging.getLogger(__name__)

class Endpoint:

    def __init__(self, url: str, query: dict, filter: dict, *args, **kwargs):
        self.url = url
        self.query = Query(**query)
        self.filter = json_parse(filter)


    def harvest(self) -> list:
        # All endpoints should extend function to do following
        # 1. prepare query
        # 2. query endpoint
        # 3. convert to stac items
        # 4. filter data if necessary
        # 5. return list of stac items

        raise NotImplementedError()

    @classmethod
    def from_config(cls, endpoint_config: dict) -> Type['Endpoint']:
        subclass_map = {subclass.type: subclass for subclass in cls.__subclasses__()}
        endpoint_type = endpoint_config.pop('type', None)
        
        SubClass = subclass_map[endpoint_type]
        return SubClass(**endpoint_config)

    def filter_data(self, data: List[dict]) -> List[dict]:
        attr_map = {
            'point_attr': 'geometry',
            '*': 'properties.*'
        }
        e = NativeEvaluator(attribute_map=attr_map, use_getattr=False)
        evaluator = e.evaluate(self.filter)
        result = list(filter(evaluator, data))
        
        return result
