import logging
from typing import List, Optional, Tuple, Type
from dataclasses import dataclass

import requests
import lxml.etree as ET
import pystac

from .Endpoint import Endpoint
from ..stac import STACItemComposer


logger = logging.getLogger(__name__)


SearchPage = Tuple[List[dict], Optional[str]]


@dataclass
class SearchPage:
    records: List[dict]
    index: int
    total: int


class OpenSearchFormat:
    mimetype = None

    @classmethod
    def from_config(cls, config: dict) -> Type['OpenSearchFormat']:
        subclass_map = {
            subclass.mimetype: subclass for subclass in cls.__subclasses__()
        }
        type_ = config.pop('type', None)

        SubClass = subclass_map[type_]
        return SubClass(**config)


class GeoJSONFormat(OpenSearchFormat):
    mimetype = 'application/json'

    def __init__(self, property_mapping):
        self.property_mapping = property_mapping

    def parse(self, response: requests.Response) -> SearchPage:
        data = response.json()
        features = [
            self._parse_feature(feature) for feature in data['features']
        ]

        return SearchPage(
            features,
            data['properties']['startIndex'],
            data['properties']['totalResults'],
        )

    def _parse_feature(self, feature: dict) -> dict:
        properties = {
            property_name: feature['properties'][property_path]
            for property_name, property_path in self.property_mapping.items()
        }
        return pystac.Item(
            feature['id'],
            geometry=feature['geometry'],
            bbox=None,
            datetime=None,
            properties=properties,
        ).to_dict()


class AtomFormat(OpenSearchFormat):
    mimetype = 'application/atom+xml'

    def __init__(self):
        # TODO: maybe add mapping from XML -> properties in output STAC
        pass


class OpenSearchEndpoint(Endpoint):
    type = 'OpenSearch'

    NS = {
        '': 'http://www.w3.org/2005/Atom',
        'opensearch': 'http://a9.com/-/spec/opensearch/1.1/',
        'parameters': 'http://a9.com/-/spec/opensearch/extensions/parameters/1.0/',
        'georss': 'http://www.georss.org/georss/',
        'media': 'http://search.yahoo.com/mrss/',
        'owc': 'http://www.opengis.net/owc/1.0/',
        'eo': 'http://a9.com/-/opensearch/extensions/eo/1.0/',
        'geo': 'http://a9.com/-/opensearch/extensions/geo/1.0/',
        'time': 'http://a9.com/-/opensearch/extensions/time/1.0/',
        'cql': 'http://a9.com/-/opensearch/extensions/cql/1.0/',
        'dc': 'http://purl.org/dc/elements/1.1/',
    }

    def __init__(self, format_config, *args, **kwargs):
        super(OpenSearchEndpoint, self).__init__(*args, **kwargs)
        self.format = OpenSearchFormat.from_config(format_config)

    def harvest(self) -> list:
        logger.info("Starting OpenSearch harvesting")

        # prepare query
        parser = ET.XMLParser(recover=True)
        data = ET.fromstring(requests.get(self.url).content, parser)
        urls = self._get_urls_params(data)
        result = []

        url, search_params, (index, index_val), _ = self.query.opensearch_params(
            urls, self.format.mimetype
        )

        while True:
            response = requests.get(url, params=search_params)
            response.raise_for_status()
            search_params[index] += index_val
            page = self.format.parse(response)
            if page.records:
                result.extend(page.records)
            else:
                break

        return result

    def _get_urls_params(self, data: ET._Element) -> dict:
        urls = {}
        for url in data.findall('opensearch:Url', self.NS):
            type = url.get('type')
            urls[type] = {}
            urls[type]['url'] = url.get('template').split('?')[0]
            urls[type]['params'] = []
            for param in url.findall('parameters:Parameter', self.NS):
                urls[type]['params'].append({
                    'name': param.get('name'),
                    'value': param.get('value'),
                    'minInclusive': param.get('minInclusive', None),
                    'maxInclusive': param.get('maxInclusive', None),
                    'minimum': param.get('minimum', None),
                    'maximum': param.get('maximum', None)
                })
        return urls
