import re
import datetime
from dateutil.parser import isoparse
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

import boto3
import botocore

from .Endpoint import Endpoint
from ..stac import STACItemComposer

class S3Endpoint(Endpoint):
    type = 'S3'

    def __init__(self, *args, **kwargs):
        super(S3Endpoint, self).__init__(*args, **kwargs)

    @property
    def client(self) -> 'S3Client':
        botocore_session = botocore.session.Session()
        session = boto3.session.Session(botocore_session=botocore_session)

        client = session.client(
            's3',
            aws_access_key_id=self.query.access_key_id,
            aws_secret_access_key=self.query.secret_access_key,
            region_name=self.query.region,
        )
        return client

    @property
    def bucket(self) -> str:
        bucket = self.url.strip('https://').split('.')[0]
        return bucket


    def harvest(self) -> list:

        # query
        paginator = self.client.get_paginator('list_objects_v2')
        pages = paginator.paginate(Bucket=self.bucket, Prefix=self.query.prefix)
        result = []

        # format as STAC
        for page in pages:
            for file in page['Contents']:
                dt = re.search(self.query.time_regex, file['Key'])[0]
                dt = isoparse(dt)
                item = STACItemComposer.from_s3(file, dt, self.url)
                result.append(item.to_dict())

        # filter
        if self.filter:
            result = self.filter_data(result)
        return result

    def _sensed_extract(self, result: list, data: dict, harvester: 'Harvester'):
        key = data['Key']
        # check if file matches regex
        if not re.search(self.file_regex, key):
            return

        # extract datetime from key
        dt = re.search(self.time_regex, key).group()
        dt = datetime.datetime.strptime(dt, self.time_format)

        # check if within given time frame
        if harvester.time_begin <= dt <= harvester.time_end:
            
            result.append(item.to_dict())

    def _modified_extract(self, result: list, data: dict, harvester: 'Harvester'):
        key = data['Key']
        # check if file matches regex
        if not re.search(self.file_regex, key):
            return

        # extract datetime from key
        dt = re.search(self.time_regex, key).group()
        dt = datetime.datetime.strptime(dt, self.time_format)

        now = datetime.datetime.now(datetime.timezone.utc)
        time_delta = datetime.timedelta(seconds=timeparse(self.time_delta))
        update_begin = now - time_delta
        if update_begin <= data['LastModified'] <= now:
            item = STACItemComposer.from_s3(data, dt, self.url)
            result.append(item.to_dict())
