from datetime import datetime, timezone
from dateutil.parser import isoparse
from typing import Tuple

from .exceptions import HarvestError

class Query:
    def __init__(self, time:dict=None, bbox:str=None, collection:str=None, *args, **kwargs):
        if time:
            self.time_begin, self.time_end, self.time_property, rest = self._time(time)
            kwargs = {** rest, **kwargs}
        else:
            self.time = time
        self.bbox = bbox
        self.collection = collection
        for k, v in kwargs.items():
            self._resolve_param(k, v)

    def _time(self, time: dict) -> Tuple[datetime, datetime, str]:

        time_begin = time.pop('begin')
        time_end = time.pop('end', None)
        time_property = time.pop('property')

        try:
            time_begin = datetime.fromordinal(time_begin.toordinal())
            time_end = datetime.fromordinal(time_end.toordinal())
        except AttributeError:
            pass

        if all([time_begin, time_end]):
            try:
                time_begin = isoparse(time_begin)
                time_end = isoparse(time_end)
            except TypeError:
                pass
        elif time_begin:
            try:
                time_end = datetime.now(timezone.utc)
                time_begin = isoparse(time_begin)
            except TypeError:
                pass
        else:
            raise HarvestError('Incorrect time parameters provided')
        
        return time_begin, time_end, time_property, time

    def _resolve_param(self, k: str, v: str):
        setattr(self, k, v)

    def stacapi_params(self, root_response: dict) -> dict:
        queryables = False
        # TODO: fetch stac queryables
        # see if queryables match configured
        for link in root_response['links']:
            if link['rel'] == 'queryables':

                queryables = link['href']

        if not queryables:
            params = {
                'datetime': f"{self.time_begin.isoformat()}/{self.time_end.isoformat()}",
                'limit': 100,
                'page': 1,
                'bbox': f"[{self.bbox}]",
                'collections': f'["{self.collection}"]'
            }
        else:
            # TODO
            params = {}

        return params

    def opensearch_params(self, urls: dict, response: str='application/atom+xml') -> Tuple[str, dict, Tuple[str, int], dict]:
        url = urls[response]['url']
        queryables = urls[response]['params']

        mapping = {
            '{geo:box}': self.bbox,
            '{time:start}': self.time_begin.isoformat(),
            '{time:end}': self.time_end.isoformat(),
            '{count}': int(next(param['maxInclusive'] for param in queryables if param["value"] == "{count}")),
            '{startIndex}': int(next(param['minInclusive'] for param in queryables if param["value"] == "{startIndex}")),
            '{geo:geometry}': None,
            '{geo:uid}': None
        }
        mapping = {k: v for k, v in mapping.items() if v is not None}

        search_params = {}
        stac_params = {}
        
        for param_name, value in mapping.items():
            for p in queryables:
                if param_name == p['value']:
                    search_params[p['name']] = value
                    if p['value'] == '{startIndex}':
                        count_name = p['name']
                    if p['value'] == '{time:start}':
                        stac_params['start_datetime'] = p['name']
                    if p['value'] == '{time:end}':
                        stac_params['end_datetime'] = p['name']
                    break


        return url, search_params, (count_name, int(mapping['{count}'])), stac_params
