import dateutil
from copy import deepcopy
from typing import TYPE_CHECKING
from datetime import timezone

from pystac import Item, Asset, Link

if TYPE_CHECKING:
    from datetime import datetime
    from lxml.etree import _Element

class STACItemComposer(Item):
    
    @classmethod
    def from_s3(cls, data: dict, dt: 'datetime', url: str) -> 'STACItemComposer':
        identifier = dt.strftime('%Y%m%d_%H%M%S')
        properties = {
            'datetime': dt.isoformat(),
            'updated': data['LastModified'].isoformat()
        }
        item = cls(id=identifier, geometry=None, bbox=None, datetime=dt, properties=properties)
        item.add_asset(identifier, Asset(f"{url}{data['Key']}"))
        return item

    @classmethod
    def from_opensearch(cls, data: '_Element', namespace: dict=None, stac_properties: dict=None) -> 'STACItemComposer':
        if namespace:
            identifier = data.find('dc:identifier', namespace).text
            link = data.find('link', namespace).get('href')
            datetime = data.find('dc:date', namespace).text
            if '/' in datetime:
                start, end = datetime.split('/')
                datetime = {
                    'start_datetime': start,
                    'end_datetime': end
                }
                item = cls(id=identifier, geometry=None, bbox=None, datetime=None, properties=datetime)
            else:
                item = cls(id=identifier, geometry=None, bbox=None, datetime=datetime, properties={})
            item.add_asset(identifier, Asset(link))
        else:
            identifier = data['id']
            datetime = {
                'start_datetime': data['properties'][stac_properties['start_datetime']],
                'end_datetime': data['properties'][stac_properties['end_datetime']]
            }
            item = cls(id=identifier, geometry=None, bbox=None, datetime=None, properties=datetime)
            asset = Asset(next(link['href'] for link in data['properties']['links'] if link['rel']=='self'))
            item.add_asset(identifier, asset)
        return item

    @classmethod
    def from_dict(cls, d, href=None, root=None):
        d = deepcopy(d)
        id = d.pop('id')
        geometry = d.pop('geometry')
        properties = d.pop('properties')
        bbox = d.pop('bbox', None)
        stac_extensions = d.get('stac_extensions')
        collection_id = d.pop('collection', None)

        datetime = properties.get('datetime')
        if datetime is not None:
            datetime = dateutil.parser.parse(datetime)
        links = d.pop('links')
        assets = d.pop('assets')

        d.pop('type')
        d.pop('stac_version')

        item = cls(id=id,
                    geometry=geometry,
                    bbox=bbox,
                    datetime=datetime,
                    properties=properties,
                    stac_extensions=stac_extensions,
                    collection=collection_id,
                    extra_fields=d)

        has_self_link = False
        for link in links:
            has_self_link |= link['rel'] == 'self'
            item.add_link(Link.from_dict(link))

        if not has_self_link and href is not None:
            item.add_link(Link.self_href(href))

        for k, v in assets.items():
            asset = Asset.from_dict(v)
            asset.set_owner(item)
            item.assets[k] = asset

        return item

    def get(self, attribute):
        return self.properties[attribute]
