import os
from os.path import basename, join, isfile
import subprocess
from typing import List
from glob import glob
import shutil
import logging
import re

from ..util import replace_ext, get_all_data_files, gdal


logger = logging.getLogger(__name__)


def calc_step(source_dir: os.PathLike, target_dir: os.PathLike, preprocessor_config: dict, formulas: List[dict], data_file_globs: List[str]=[]):
    for i, item in enumerate(formulas):
        # get all filenames, first looking into source_dir, then target_dir as fallback
        filenames_source = glob(join(source_dir, list(item['inputs'].values())[0]['glob']))
        if len(filenames_source) > 0:
            filenames_final = filenames_source
        else:
            filenames_target = glob(join(target_dir, list(item['inputs'].values())[0]['glob']))
            if len(filenames_target) > 0:
                filenames_final = filenames_target
            else:
                raise Exception('No input file in source or target directory for calc: %s' % item)
        for filename in filenames_final:
            target_filename = join(
                target_dir,
                replace_ext(basename(filename), item.get('output_postfix', '_proc%d' % i) + '.tif', False)
            )

            if isfile(target_filename):
                logger.warn('Calc output filename %s already exists' % target_filename)

            calc_formula(source_dir, target_dir, item['inputs'], target_filename, item['formula'], item.get('data_type', 'Float32'), item.get('nodata_value', None), item.get('creationOptions', {}))

    # take all original files with from the last step matching the data_file_glob
    filenames = get_all_data_files(source_dir, preprocessor_config, data_file_globs)
    for filename in filenames:
        target_filename = join(target_dir, basename(filename))
        if isfile(target_filename):
            logger.warn('Calc output filename %s already exists' % target_filename)
        shutil.copy(filename, target_filename)


def calc_formula(source_dir: os.PathLike, target_dir: os.PathLike, inputs: List[dict], target_filename: os.PathLike, formula: str, data_type: str="Float32", nodata_value: float=None, creationOptions: List[str]=[]):
    used_formula = formula
    cmd = [
        "gdal_calc.py",
        "--outfile=%s" % target_filename,
        "--type", data_type,
    ]
    for input_name in inputs:
        # get first filename as a base, first looking into source_dir, then target_dir as fallback
        filenames_source = glob(join(source_dir, inputs[input_name]['glob']))
        if len(filenames_source) > 0:
            filename = filenames_source[0]
        else:
            filenames_target = glob(join(target_dir, inputs[input_name]['glob']))
            if len(filenames_target) > 0:
                filename = filenames_target[0]
            else:
                raise Exception('No input file in source or target directory for formula: %s' % formula)
        band_number = inputs[input_name].get('band', 1)
        cmd.extend([
            "-%s" % input_name, filename,
            "--%s_band=%d" % (input_name, band_number),
        ])
        # evaluate formula as a template
        used_formula = evaluate_formula(formula, filename, input_name, band_number)

    cmd.extend([
        "--calc=%s" % used_formula,
    ])

    if nodata_value is not None:
        cmd.extend([
            "--NoDataValue=%s" % nodata_value,
        ])

    for option in creationOptions:
        cmd.extend([
            "--co",
            option,
        ])

    process = subprocess.run(cmd, capture_output=True, text=True)
    logger.debug("gdal calc stderr: %s" % process.stderr)


def evaluate_formula(formula, filename, input_name, band):
    """ Tries to replace a few common placeholders in the calc formula
    """
    # formula: "10*log10((A.astype(float)-${A_1_statistics_max}+0.0001)/(${A_1_statistics_max}-${A_1_statistics_min}))"
    evaluated_formula = formula
    if "statistics" in formula:
        # find all occurence of templates in form of ${A_1_statistics_max}
        found = re.findall(r'(\${[0-9a-zA-Z_]*})', formula)
        # get unique, keeping order
        found = list(dict.fromkeys(found))
        for item in found:
            parts = item.replace("${", "").replace("}", "").split("_")
            # if current input is the one in template, replace it with statistics
            if parts[0] == input_name and int(parts[1]) == band:
                src_ds = gdal.Open(filename)
                statistics = src_ds.GetRasterBand(band).GetStatistics(True, True)  # force recount
                del src_ds
                if parts[3] == "min":
                    replace = "%s" % statistics[0]
                elif parts[3] == "max":
                    replace = "%s" % statistics[1]
                elif parts[3] == "mean":
                    replace = "%s" % statistics[2]
                elif parts[3] == "std":
                    replace = "%s" % statistics[3]
                else:
                    logger.warn("Unknown statistics found in expression %s" % item)
                    continue
                # replace the template with actual value
                evaluated_formula = evaluated_formula.replace(item, replace)
    return evaluated_formula
